"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

# -*- coding: utf-8 -*-
import uuid, sys, os.path
from node import Node
from backup_and_restore_manager import BackupAndRestoreManager
from libs import base
from libs import citrusleaf as cl
from constants import *
from datetime import datetime
from alerts import cluster_level
from utilities import alert_manager
from utilities import connection_manager
from helpers.seed_down_handler import HandleSeedDown
from helpers.seed_down_handler_with_version import HandleSeedDownWithVersion
from helpers import list_comparator
from helpers import ip_checker
from constants import ROLES
from models.user import User
from werkzeug.exceptions import Unauthorized
from operator import itemgetter

"""ERROR_DESCRIPTION = {1: "Unknown Server Error",
     50: "Last 'proto' in a Chunked Sequence",
     51: "Security Not Supported",
     52: "Security Not Enabled",
     53: "Security Scheme Not Supported",
     54: "Unrecognized Command ID",
     55: "Unrecognized Field ID",
     56: "Illegal State (e.g. valid but unexpected command)",
     60: "No User or Unrecognized User",
     61: "User Already Exists",
     62: "No Password or Bad Password",
     65: "No Credential or Bad Credential",
     70: "No Role(s) or Unrecognized Role(s)",
     80: "Not Authenticated",
     81: "Role/Privilege Violation"}"""

class Cluster(object):
    """Cluster contains all information related to cluster"""

    def __new__(cls, ip, port, user=None, password=None):

        if base.is_node_available(ip, port, user=user, password=password) is False:
            return UNABLE_TO_CONNECT
        return super(Cluster, cls).__new__(cls, ip, port, user=user, password=password)

    def __init__(self, ip, port, user=None, password=None):

        self._ip = ip
        self._port = port
        self._seed = "%s:%s" % (ip, port)
        self._cluster_id = str(uuid.uuid4())
        self._paused_state = False

        if user and password:
            self._users = {
                user: {
                    "username": user,
                    "password": password,
                    "roles": []
                    }
                }
        else:
            self._users = {}

        self._security = base.is_security_enabled(self._ip, self._port)
        self.update_user_roles(user)
        self._nodes = {self._seed: Node(self._seed, self._cluster_id, user=user, password=password)}
        self._last_access_time = datetime.now()
        self._alert_manager = alert_manager.AlertManager()

        self._id = None
        self._node_list = set([self._seed])
        self._off_nodes = set()
        self._current_nodes = []
        self._index_list = {}
        self._node_visibilties = {}
        self.get_updates(user=user, password=password)
        self.set_status_on()
        self._latest_build_no = self.get_build_no_of_cluster_nodes().get("latest_build_no")
        self._namespaces = self.get_namespaces(user=user, password=password)
        self._cluster_key = self._nodes[self._seed].get_cluster_key(user=user, password=password)

        self._alert = cluster_level.ClusterLevel(self._cluster_id)
        self._backup_and_restore_manager = BackupAndRestoreManager(self._ip, self._port)

    def get_cluster_id(self):

        return self._cluster_id

    def get_updates(self, alerts_flag=False, user=None, password=None):

        address = "%s:%s" % (self._ip, self._port)
        node_obj = self.get_node(address)
        if node_obj is not None:
            sock = node_obj.get_socket_from_queue()
            response = base.check_authenticity_of_user(self._ip, self._port, user, password, sock=sock)
        else:
            response = base.check_authenticity_of_user(self._ip, self._port, user, password)
        if type(response) is int:
            new_sock = connection_manager.create_socket(self._ip, self._port, user=user, password=password)
            if new_sock is not None:
                sock = new_sock
                response = base.check_authenticity_of_user(self._ip, self._port, user, password, sock=sock)
        if node_obj is not None:
            node_obj.put_socket_in_queue(sock)
        if response in [UNRECOGNISED_USER, BAD_PASSWORD, BAD_CREDENTIAL]:
            raise Unauthorized('Unauthenticated user!')
        #if user and password and base.check_authenticity_of_user(self._ip, self._port,
        #                                       user, password, sock=sock) in [UNRECOGNISED_USER, BAD_PASSWORD, BAD_CREDENTIAL]:
         #   raise Unauthorized("Unauthenticated user!")
        #if user and password and base.check_authenticity_of_user(self._ip, self._port,
         #                                      user, password) in [UNRECOGNISED_USER, BAD_PASSWORD, BAD_CREDENTIAL]:
         #   raise Unauthorized("Unauthenticated user!")

        enterprise_flag = os.path.isfile(ENTERPRISE_FLAG_FILE)
        node_list = set()
        if enterprise_flag and alerts_flag:
            cluster_integrity = "true"
            is_cluster_size_same = True
            node_cluster_size = None

        for node in list(self._nodes.values()):

            if not node.is_valid():
                continue
            node_list = node_list.union(node.get_node_list(user=user, password=password))

            if enterprise_flag and alerts_flag:
                if node.get_statistic("cluster_integrity") == "false":
                    cluster_integrity = "false"
                current_node_cluster_size = node.get_statistic("cluster_size")
                if current_node_cluster_size != NOT_SUPPORTED:
                    if node_cluster_size is None:
                        node_cluster_size = current_node_cluster_size
                    elif node_cluster_size != current_node_cluster_size:
                        is_cluster_size_same = False

        if enterprise_flag and alerts_flag:
            self._alert.on_cluster_integrity_change(cluster_integrity)
            self._alert.on_cluster_size_change(is_cluster_size_same)
        new_node_list = node_list - self._node_list

        for node_address in new_node_list:
            actual_node_address = self.add_new_node(node_address,
                                                user=user, password=password)
            if actual_node_address is None:

                node_list.remove(node_address)
            elif enterprise_flag and alerts_flag:
                self._alert.on_new_node_add(actual_node_address)


        self.calculate_visibilities(user=user, password=password)

    def add_existing_node(self, node_address, user=None, password=None):

        ip, port = node_address.split(":")
        ip = ip_checker.check_ip(ip)

        node_address = "%s:%s" % (ip, port)

        for address in self._node_list:
            if address.startswith(ip + ":"):
                return dict(status="failure", error="Node already exists")

        if ip == "Unable to connect":
            return dict(status="failure", error="Unable to connect")
        elif ip == "Unable to resolve hostname":
            return dict(status="failure", error="Unable to resolve hostname")
        elif ip == "Please enter network IP":
            return dict(status="failure", error="Please enter network IP")
        else:
            self._node_list.add(node_address)
            self._nodes[node_address] = Node(node_address, self._cluster_id, validity=None, user=user, password=password)
            if self._nodes[node_address].get_status(user=user, password=password) == "on" and self.get_cluster_key(user=user, password=password) != self._nodes[node_address].get_cluster_key(user=user, password=password):
                self.remove_node(node_address)
                return dict(status="Node <strong>%s</strong> cannot be monitored here as it belongs to a different cluster." % node_address)
            else:
                return dict(status="success")

    def add_new_node(self, node_address, user=None, password=None):

        ip, port = node_address.split(":")
        node_address = base.get_self_address(ip, port, user=user, password=password)
        if node_address in cl.ERROR_CODES or node_address in self._node_list:
            return None
        self._node_list.add(node_address)
        self._nodes[node_address] = Node(node_address, self._cluster_id, user=user, password=password)
        return node_address

    def remove_node(self, node_address):

        if node_address in self._node_list and node_address in self._nodes:
            self._alert_manager.delete_alerts_specific_to_node(node_address)
            self._node_list.remove(node_address)
            if node_address in self._off_nodes:
                self._off_nodes.remove(node_address)
            del self._nodes[node_address]
            return dict(status="success")
        return dict(status="failure")

    def get_id(self):
        return self._id

    def get_seed_ip(self):
        return self._ip

    def get_seed_port(self):
        return self._port

    def get_seed(self):
        return self._seed

    def set_seed(self, seed):
        self._seed = seed
        self._ip, self._port = seed.split(":")

    def get_status(self):
        return self._status

    def set_status_off(self):
        self._status = "off"

    def set_status_on(self):
        self._status = "on"

    def get_last_access_time(self):
        return self._last_access_time

    def update_last_access_time(self):
        self._last_access_time = datetime.now()

    def get_nodes(self):
        return self._nodes

    def get_user(self):
        return self._users

    def get_password(self):
        return self.password

    def is_secured(self):
        return self._security

    @HandleSeedDown()
    def get_build_no(self, ip=None, port=None, user=None, password=None):

        ip = ip or self._ip
        port = port or self._port

        address = "%s:%s" % (self._ip, self._port)
        node_obj = self.get_node(address)

        if node_obj is not None:
            sock = node_obj.get_socket_from_queue()
            build_no = base.get_node_build(self._ip, self._port, user=user, password=password, sock=sock)
        else:
            build_no = base.get_node_build(self._ip, self._port, user=user, password=password)

        if build_no is None:
            new_sock = connection_manager.create_socket(self._ip, self._port, user=user, password=password)
            if new_sock is not None:
                sock = new_sock
                build_no = base.get_node_build(self._ip, self._port, user=user, password=password, sock=sock)
        if node_obj is not None:
            node_obj.put_socket_in_queue(sock)
        #build_no = base.get_node_build(self._ip, self._port, user=user, password=password)

        return build_no
        #return base.get_node_build(self._ip, self._port, user=user, password=password)

    def get_build_no_of_cluster_nodes(self):

        info = dict(version_list=dict())
        for node in list(self._nodes.values()):
            build_no = node.get_build()
            if build_no is not None:
                if build_no in list(info['version_list'].keys()):
                    info['version_list'][build_no].append(node.get_address())
                else:
                    info['version_list'][build_no] = [node.get_address()]
        build_list = []
        for build_no in list(info['version_list'].keys()):
            build_list.append(build_no.replace('.', ''))
        if list(info['version_list'].keys()) == []:
            info['latest_build_no'] = "Undefined"
        else:
            info['latest_build_no'] = list(info['version_list'].keys())[build_list.index(max(build_list))]
        return info

    def get_cluser_compatibility_details(self):

        info = dict()
        info['build_details'] = self.get_build_no_of_cluster_nodes()
        version_list = info['build_details']['version_list']
        
        if len(version_list) == 0 or len(version_list) == 1:
            info['nodes_compatibility'] = 'homogeneous'
        else:
            min_v = max_v = version_list.keys()[0]
            flag = True
            for key in version_list.keys():
                if key < min_v:
                    min_v = key
                    flag = False
                elif key > max_v:
                    max_v = key
                    flag = False
            if flag is True:
                info['nodes_compatibility'] = 'homogeneous'
            else:
                min_v = int(str(min_v).replace(".", ""))
                max_v = int(str(max_v).replace(".", ""))
                if min_v < 354 and max_v >= 354:
                    info['nodes_compatibility'] = 'incompatible'
                else:
                    info['nodes_compatibility'] = 'compatible'
        return info['nodes_compatibility']

    def get_latest_build_no(self):

        return self._latest_build_no

    def get_node_list(self):

        return list(self._node_list)

    def get_off_nodes(self, user=None, password=None):

        off_nodes = []
        for node in list(self._nodes.values()):
            if node.get_status(user=user, password=password) == "off":
                off_nodes.append(node.get_address())
        return off_nodes

    def get_different_cluster_nodes(self):

        different_cluster_nodes = []
        for node in list(self._nodes.values()):
            if node.is_valid() is False:
                different_cluster_nodes.append(node.get_address())
                self.remove_node(node.get_address())
        return different_cluster_nodes

    def get_number_of_nodes(self):

        return len(self._node_list)

    def get_namespaces(self, user=None, password=None):

        self._namespaces = self.update_namespaces(user=user, password=password)
        return self._namespaces

    def get_number_of_namespaces(self, user=None, password=None):

        return len(self.get_namespaces(user=user, password=password))

    def update_namespaces(self, user=None, password=None):

        namespace_list = set()
        for node in list(self._nodes.values()):
            namespace_list = namespace_list.union(node.get_namespace_list(user=user, password=password))
        return list(namespace_list)

    def get_memory_usage(self):

        total_bytes_memory = 0
        used_bytes_memory = 0
        details = dict()

        for node_obj in list(self._nodes.values()):
            if node_obj is None or not node_obj.is_valid():
                continue

            used, total = node_obj.get_memory()
            if used.isdigit() and total.isdigit():
                total_bytes_memory += int(total)
                used_bytes_memory += int(used)
                details[node_obj.get_address()] = dict(used=int(used), free=int(total)-int(used))

        total_free_memory = total_bytes_memory - used_bytes_memory

        return dict(used=used_bytes_memory, free=total_free_memory, details=details)

    def get_disk_usage(self):

        total_bytes_disk = 0
        used_bytes_disk = 0
        details = dict()

        for node_obj in list(self._nodes.values()):
            if node_obj is None or not node_obj.is_valid():
                continue

            used, total = node_obj.get_disk()
            if used.isdigit() and total.isdigit():
                total_bytes_disk += int(total)
                used_bytes_disk += int(used)
                details[node_obj.get_address()] = dict(used=int(used), free=int(total)-int(used))

        total_free_disk = total_bytes_disk - used_bytes_disk

        return dict(used=used_bytes_disk, free=total_free_disk, details=details)

    def get_latencies(self):

        self._latencies = []
        for node in list(self._nodes.values()):
            latency = dict(address=node.get_address())
            latency.update(node.get_latency())
            self._latencies.append(latency)

        return self._latencies

    def get_throughput(self, node_addresses, user=None, password=None):

        throughput = dict()
        for node_address in node_addresses:
            node = self.get_node(node_address)
            if node is None or node.get_status(user=user, password=password) == "off" or not node_obj.is_valid():
                continue
            throughput[node_address] = node.get_throughput()
        return throughput

    def get_statistics(self, user=None, password=None):

        self._statistics = []
        for node in list(self._nodes.values()):
            if node.is_valid():
                stat = dict(address=node.get_address())
                stat.update(node.get_updated_statistics(self, user=user, password=password))
                self._statistics.append(stat)
        return self._statistics

    def update_statistics(self, user=None, password=None):

        for node in list(self._nodes.values()):
            if node.is_valid() is not False:
                node.get_updated_statistics(self, user=user, password=password)

    def update_latencies(self, user=None, password=None):

        for node in list(self._nodes.values()):
            if node.is_valid() is not False:
                node.get_updated_latencies(user=user, password=password)

    def calculate_visibilities(self, user=None, password=None):

        cur_nodes = set()

        for node_obj in list(self._nodes.values()):

            #if not node_obj.is_valid():
            if node_obj.is_valid() is False:
                continue
            nodes = node_obj.get_node_list(user=user, password=password)

            cur_nodes = cur_nodes.union(set(nodes))


            old_visibility = node_obj.get_cluster_visibility()
            visibility = True

            for node2_obj in list(self._nodes.values()):

                if not node2_obj.is_valid() or node_obj.get_address() == node2_obj.get_address():
                    continue

                nodes2 = node2_obj.get_node_list(user=user, password=password)
                if nodes2 == set():
                    continue

                if node_obj.get_address() in nodes2:
                    visibility &= True
                elif node_obj.get_address() not in nodes2 and nodes != set():
                    visibility = False
                    break
            if old_visibility != visibility:
                node_obj.set_cluster_visibility(visibility)


        cur_nodes = cur_nodes.intersection(set(self.get_node_list()))
        off_nodes = set(self.get_node_list()) - cur_nodes

        for node in (off_nodes-self._off_nodes):

            self._nodes[node].set_status("off")
            #self._nodes[node].set_invalid()
        for node in (self._off_nodes-off_nodes):
            self._nodes[node].set_status("on")
            #self._nodes[node].set_valid()
        self._off_nodes = off_nodes

    def get_node_visibility(self, node_address):

        return self._node_visibilties.get(node_address, None)

    def get_node(self, node_address):

        return self._nodes.get(node_address, None)

    def get_namespace(self, namespace_name, node_address):

        node_obj = self.get_node(node_address)
        if node_obj is None or node_obj.is_valid() is False:
            return None
        return node_obj.get_namespace(namespace_name)

    def get_xdr(self, xdr_port, node_address, user=None, password=None):

        node_obj = self.get_node(node_address)
        if node_obj is None or node_obj.is_valid() is False:
            return None
        return node_obj.get_xdr(xdr_port, user=user, password=password)

    def get_dc_names(self, xdr_port="3004", user=None, password=None):

        dc_names = dict()
        for node in list(self._nodes.values()):
            address = node.get_address()
            xdr = self.get_xdr(xdr_port, address,
                                user=user, password=password)
            if xdr is None:
                dc_names[address] = "node off"
            elif xdr == {}:
                dc_names[address] = "xdr off"
            else:
                dc_names[address] = xdr.get_dc_names(user=user, password=password)
        return dc_names

    def get_alerts(self, last_id=0, user=None, password=None):

        address = "%s:%s" % (self._ip, self._port)
        node_obj = self.get_node(address)

        if node_obj is not None:
            sock = node_obj.get_socket_from_queue()
            response = base.check_authenticity_of_user(self._ip, self._port, user, password, sock=sock)
        else:
            response = base.check_authenticity_of_user(self._ip, self._port, user, password)
        if type(response) is int:
            new_sock = connection_manager.create_socket(self._ip, self._port, user=user, password=password)
            if new_sock is not None:
                sock = new_sock
                response = base.check_authenticity_of_user(self._ip, self._port, user, password, sock=sock)
        if node_obj is not None:
            node_obj.put_socket_in_queue(sock)
        if response in [UNRECOGNISED_USER, BAD_PASSWORD, BAD_CREDENTIAL]:
            raise Unauthorized('Unauthenticated user!')
        #if user and password and base.check_authenticity_of_user(self._ip, self._port,
        #                                       user, password, sock=sock) in [UNRECOGNISED_USER, BAD_PASSWORD, BAD_CREDENTIAL]:
         #   raise Unauthorized("Unauthenticated user!")
        #if user and password and base.check_authenticity_of_user(self._ip, self._port,
        #                                       user, password) in [UNRECOGNISED_USER, BAD_PASSWORD, BAD_CREDENTIAL]:
        #    raise Unauthorized("Unauthenticated user!")

        self.update_alerts(user=user, password=password)
        return self._alert_manager.get_alerts(self._cluster_id, last_id)

    def get_active_red_alert_count(self, user=None, password=None):

        self.update_alerts(user=user, password=password)
        return self._alert_manager.get_active_red_alert_count(self._cluster_id)

    def update_alerts(self, user=None, password=None):

        for node in list(self._nodes.values()):
            node.update_alerts(user=user, password=password)

    def delete_alerts(self):

        self._alert_manager.delete_alerts(self._cluster_id)

    def delete_old_alerts(self):

        self._alert_manager.delete_old_alerts(self._cluster_id)

    @HandleSeedDownWithVersion()
    def get_indexes(self, ip=None, port=None, user=None, password=None):

        ip = ip or self._ip
        port = port or self._port
        address = "%s:%s" % (ip, port)
        node_obj = self.get_node(address)
        
        try:
            if node_obj is not None:
                sock = node_obj.get_socket_from_queue()
                result = base.get_indexes(ip, port, user=user, password=password, sock=sock)
            else:
                result = base.get_indexes(ip, port, user=user, password=password)

            if type(result) is int:
                new_sock = connection_manager.create_socket(ip, port, user=user, password=password)
                if new_sock is not None:
                    sock=new_sock
                    result = base.get_indexes(ip, port, user=user, password=password, sock=sock)
        except:
            result = UNSUPPORTED_VERSION
        if node_obj is not None:
            node_obj.put_socket_in_queue(sock)
        return result

    def get_indexes_by_namespaces(self, namespace, user=None, password=None):

        indexes = self.get_indexes(user=user, password=password)
        if indexes == UNABLE_TO_CONNECT:
            return UNABLE_TO_CONNECT
        elif indexes == UNSUPPORTED_VERSION:
            indexes = {}

        return indexes.get(namespace, [])

    def get_index_list_by_namespace(self, user=None, password=None):

        index_details = self.get_indexes(user=user, password=password)
        if index_details == UNABLE_TO_CONNECT:
            return {}
        elif index_details == UNSUPPORTED_VERSION:
            index_details = {}

        index_list = {}
        for namespace, indexes in list(index_details.items()):
            index_list[namespace] = [index["indexname"] for index in indexes]
        return index_list

    @HandleSeedDown()
    def get_sets_info(self, ip=None, port=None, user=None, password=None):

        first_iteration = True
        previous_sets_info = None
        for node_obj in list(self._nodes.values()):
            address = node_obj.get_address()
            ip, port = str(address).split(':')
            sock = node_obj.get_socket_from_queue()
            sets_info = base.get_sets_info(ip, port, sock=sock)
            if type(sets_info) is int:
                new_sock = connection_manager.create_socket(ip, port)
                if new_sock is not None:
                    sock = new_sock
                    sets_info = base.get_sets_info(ip, port, sock=sock)
            node_obj.put_socket_in_queue(sock)
            if sets_info != {} and sets_info != UNABLE_TO_CONNECT:
                if first_iteration is False:
                    for key in sets_info.iterkeys():
                        sets = sets_info[key]
                        for set_obj in sets:
                            previous_sets = previous_sets_info[key]
                            try:
                                set_index = map(itemgetter('set_name'),previous_sets).index(set_obj['set_name'])
                            except ValueError:
                                previous_sets.append(set_obj)
                                continue
                            #set_index = previous_sets.index(set_obj)
                            previous_set = previous_sets[set_index]
                            previous_set['n_objects'] =  int(previous_set['n_objects']) + int(set_obj['n_objects'])
                else:
                    first_iteration = False
                    previous_sets_info = sets_info
            #else:
                #previous_sets_info = sets_info
        # Flow will enter inside if there are no sets present.
        if previous_sets_info is None:
            previous_sets_info = sets_info
        return previous_sets_info

    def get_sets_info_by_namespaces(self, namespace,
                                    user=None, password=None):

        sets_info = self.get_sets_info(user=user, password=password)
        if sets_info == UNABLE_TO_CONNECT:
            return UNABLE_TO_CONNECT

        return sets_info.get(namespace, [])

    @HandleSeedDown()
    def get_cluster_key(self, ip=None, port=None, user=None, password=None):

        try:
            self._cluster_key = self._nodes[self._seed].get_cluster_key(user=user, password=password)

        except KeyError:
            return None
        return self._cluster_key

    def initiate_backup(self, namespace, destination_node_address, username,
    password, destination_location, sets, only_metadata,
    scan_priority, terminate_on_change, usr=None, passwd=None, pl_password=None):

        return self._backup_and_restore_manager.initiate_backup(namespace,
        destination_node_address, username, password, destination_location,
        self.get_build_no(user=usr, password=passwd), sets, only_metadata, scan_priority,
        terminate_on_change, usr=usr, passwd=passwd, pl_password=pl_password)

    def get_backup_progress(self):

        return self._backup_and_restore_manager.get_backup_progress(self._alert)

    def get_successful_backups(self):

        return self._backup_and_restore_manager.get_successful_backups_initiated_from_amc()

    def get_available_backups(self, source_node_address, source_location,
    username, password):

        return self._backup_and_restore_manager.get_available_backups(
        source_node_address, source_location, username, password)

    def initiate_restore(self, source_node_address, source_location, threads,
    namespace, missing_records_only, ignore_generation_num, username, password, usr=None, passwd=None, pl_password=None):

        return self._backup_and_restore_manager.initiate_restore(
        source_node_address, source_location, username, password,
        self.get_build_no(user=usr, password=passwd), threads, namespace, missing_records_only,
        ignore_generation_num, usr=usr, passwd=passwd, pl_password=pl_password)

    def get_restore_progress(self, user=None, password=None):

        return self._backup_and_restore_manager.get_restore_progress(
        self.get_build_no(user=user, password=password), self._alert)

    # All methods related to User dict maintained in a cluster object
    def update_user_roles(self, user):

        user_info = self._users.get(user, {})
        if user_info:
            user_obj = User(user_info.get("username"),
                            user_info.get("password"),
                            host=self._seed)
            user_info["roles"] = user_obj.get_roles(str(self._ip), str(self._port))
        #else:
           # sys.stderr.write("No user with username :" + str(user))

    def add_user(self, user_obj):

        if user_obj:
            user_info = {
                "username": user_obj.get_id(),
                "password": user_obj.get_password(),
                "roles": user_obj.get_roles(str(self._ip), str(self._port))
                }
            self._users[user_obj.get_id()] = user_info

    def remove_user(self, user_obj):

        try:
            if user_obj:
                self._users.pop(user_obj.get_id())
        except KeyError:
            import sys
            #sys.stderr.write("User:" + str(user_obj.get_id()) + "does not present on this cluster.")

    def get_highest_role_user(self):

        #build = base.get_node_build(ip, port, user=user, password=password)
        return User(self._users.values()[0]['username'], self._users.values()[0]['password'], host=self._seed)
        """
        all_users = self._users
        highest_role_user = None
        highest_role = -1
        for user, user_info in all_users.iteritems():

            role_vals = [ROLES[role] for role in user_info['roles']]
            if role_vals and max(role_vals) > highest_role:
                highest_role = max(role_vals)
                highest_role_user = User(
                                        user_info.get("username"),
                                        user_info.get("password"),
                                        host = self._seed)
        return highest_role_user"""

    def authenticate_user(self, user):

        return user in self._users.keys()

    def pause(self):

        if self._paused_state == False:
            for node in list(self._nodes.values()):
                node.append_on_pause_tps_history()
                node.append_on_pause_latency_history()

        self._paused_state = True

    def resume(self):

        if self._paused_state == True:
            for node in list(self._nodes.values()):
                node.append_on_resume_tps_history()
                node.append_on_resume_latency_history()

        self._paused_state = False

    def is_paused(self):
        return self._paused_state

    def change_node_password(self, user, password):

        for node in list(self._nodes.values()):
            node.change_password(password)
            node.replace_socket_queue()
            xdr_obj = node.get_xdr_obj()
            if hasattr(xdr_obj, '_ip'):
                xdr_obj.replace_socket_queue(user, password)

    def add_new_user(self, user, password, roles):

        if roles is None:
            status, error_code = base.add_user(self, user, password)
        else:
            status, error_code = base.add_user(self, user, password, roles)
        if status:
            return dict(status="success")
        return dict(status="failure", error=ERROR_DESCRIPTION[error_code])

    def drop_user(self, user):

        status, error_code = base.drop_user(self, user)
        if status:
            return dict(status="success")
        return dict(status="failure", error=ERROR_DESCRIPTION[error_code])

    def update_user(self, user, roles, old_roles, password):

        if user in self._users:
            return dict(status="failure", error="Current User cannot be updated!!"), False
        if roles is not None:
            if self._latest_build_no < '3.5.4':
                roles_status, error_code_roles = base.replace_roles(self, user, roles)
            else:
                resp = list_comparator.compare_roles(roles, old_roles)
                added_roles = resp['added_roles']
                revoked_roles = resp['deleted_roles']
                if added_roles:
                    #add roles to user
                    added_roles_status, error_code_added_roles = base.grant_roles(self, user, added_roles)
                else:
                    added_roles_status = True
                if revoked_roles:
                    #revoke roles from user
                    revoked_roles_status, error_code_revoked_roles = base.revoke_roles(self, user, revoked_roles)
                else:
                    revoked_roles_status = True
                roles_status = added_roles_status and revoked_roles_status
        else:
            roles_status = True
        if password is not None:
            pwd_status, error_code_pwd = base.set_password(self, user, password)
            status = roles_status and pwd_status
        else:
            pwd_status = True
            status = roles_status
        import collections
        compare = lambda x, y: collections.Counter(x) == collections.Counter(y)
        if status is False:
            if roles_status == pwd_status:
                error_msg = "Roles update failed due to: " + str(ERROR_DESCRIPTION[error_code_roles])+" & password update failed due to: " + str(ERROR_DESCRIPTION[error_code_pwd])
                return dict(status="failure", error=error_msg), False
            else:
                if roles_status:
                    error_msg = "Updating roles operation successful, password operation failed due to: " + str(ERROR_DESCRIPTION[error_code_pwd])
                    if len(roles) <=  len(old_roles) and list_comparator.compare(roles, old_roles) is False:
                        return dict(status="failure", error=error_msg), True
                    return dict(status="failure", error=error_msg), False
                else:
                    error_msg = "Updating password operation successful, roles operation failed due to: " + str(ERROR_DESCRIPTION[error_code_roles])
                    return dict(status="failure", error=error_msg), True
        #if user in self._users:
            #sys.stderr.write('\nCurrent USer changed...')


        if (len(roles) <=  len(old_roles) and compare(old_roles, roles) is False) or password is not None:
            return dict(status="success"), True
        return dict(status="success"), False

    def get_all_users(self):

        response = base.get_all_users(self)
        status = response[0]
        users = response[1]
        queryRoles_status, all_roles = base.query_roles(self)
        if type(status) == int and status not in cl.ERROR_CODES:
            users_list = []
            for user in users:
                temp_dict = {}
                temp_dict['user'] = user.name
                temp_dict['roles'] = user.roles
                users_list.append(temp_dict)
            return dict(status="success", users=users_list, roles=all_roles)
        if status is -1:
            return dict(status="failure", users=[], error="Unknown Reason", roles=None)
        return dict(status="failure", users=[], error=ERROR_DESCRIPTION[status], roles=None)

    def change_password(self, user, old_password, new_password):

        if str(self._users[user]['password']) != str(old_password):
            return dict(status="failure", error="Invalid current password")
        if str(new_password) == str(old_password):
            return dict(status="failure", error="New password cannot be same as current password")
        status, error_code = base.change_password(self, user, old_password, new_password)
        if status:
            self._users[user]['password'] = str(new_password)
            self.change_node_password(str(user), str(new_password))
            return dict(status="success")
        return dict(status="failure", error=ERROR_DESCRIPTION[error_code])

    def close_sockets(self):

        for node in list(self._nodes.values()):
            node.close_sockets()

    def get_all_roles(self):

        status, response = base.query_roles(self)
        if status == 0:
            return dict(status='success', roles=response)
        else:
            if status in ERROR_DESCRIPTION:
                error_msg = ERROR_DESCRIPTION[status]
            else:
                error_msg = 'Unknown Error... Server Returned: ' + str(status)
            return dict(status='failure', error=error_msg, roles=None)
        return response

    def get_role(self, role):

        address = "%s:%s" % (self._ip, self._port)
        node_obj = self.get_node(address)
        user = node_obj.user
        password = node_obj.password
        status, privileges = base.query_role(self._ip, self._port, user, password, str(role))
        if status == 0 :
            return dict(status='success', privileges=privileges)
        err_msg = ERROR_DESCRIPTION[status] if status in ERROR_DESCRIPTION else 'Unknown Error..Server Returned: ' + str(status)
        return dict(status='failure', privileges=None, error=err_msg)

    def add_role(self, role, privileges):

        status = base.create_role(self, role, privileges)
        if status == 0:
            return dict(status='success')
        err_msg = ERROR_DESCRIPTION[status] if status in ERROR_DESCRIPTION else 'Unknown Error..Server Returned: ' + str(status) 
        return dict(status='failure', error=err_msg)

    def drop_role(self, role):

        status = base.delete_role(self, role)
        if status == 0:
            return dict(status='success')
        err_msg = ERROR_DESCRIPTION[status] if status in ERROR_DESCRIPTION else 'Unknown Error..Server Returned: ' + str(status) 
        return dict(status='failure', error=err_msg)

    def add_privileges(self, role, privileges):

        status = base.add_privileges(self, role, privileges)
        if status == 0:
            return dict(status='success')
        err_msg = ERROR_DESCRIPTION[status] if status in ERROR_DESCRIPTION else 'Unknown Error..Server Returned: ' + str(status) 
        return dict(status='failure', error=err_msg)

    def drop_privileges(self, role, privileges):

        status = base.drop_privileges(self, role, privileges)
        if status == 0:
            return dict(status='success')
        err_msg = ERROR_DESCRIPTION[status] if status in ERROR_DESCRIPTION else 'Unknown Error..Server Returned: ' + str(status) 
        return dict(status='failure', error=err_msg)

    def update_role(self, role, privileges, old_privileges):

        resp = list_comparator.compare_roles(privileges, old_privileges)
        added_privileges = resp['added_roles']
        revoked_privileges = resp['deleted_roles']
        if added_privileges:
            #add privileges to user
            added_privileges_status = base.add_privileges(self, role, added_privileges)
        else:
            added_privileges_status = 0
        if revoked_privileges:
            #revoke privileges from user
            revoked_privileges_status = base.drop_privileges(self, role, revoked_privileges)
        else:
            revoked_privileges_status = 0
        
        if added_privileges_status == 0 and revoked_privileges_status == 0:
            return dict(status='success')
        else:
            if added_privileges_status != 0 and revoked_privileges_status != 0:
                added_prvileges_err = ERROR_DESCRIPTION[added_privileges_status] if added_privileges_status in ERROR_DESCRIPTION else 'Unknown Error..Server Returned: ' + str(added_privileges_status) 
                revoked_privileges_err = ERROR_DESCRIPTION[revoked_privileges_status] if revoked_privileges_status in ERROR_DESCRIPTION else 'Unknown Error..Server Returned: ' + str(revoked_privileges_status) 
                err_msg = 'Grant Privileges failed due to: %s and Revoke Privileges failed due to: %s' %( str(added_prvileges_err), str(revoked_privileges_err) )
                return dict(status='failure', error= err_msg)
            else:
                if added_privileges_status != 0:
                    added_prvileges_err = ERROR_DESCRIPTION[added_privileges_status] if added_privileges_status in ERROR_DESCRIPTION else 'Unknown Error..Server Returned: ' + str(added_privileges_status)         
                    err_msg = 'Grant Privileges failed due to: %s' %( str(added_prvileges_err))
                    return dict(status='failure', error= err_msg)
                else:
                    revoked_privileges_err = ERROR_DESCRIPTION[revoked_privileges_status] if revoked_privileges_status in ERROR_DESCRIPTION else 'Unknown Error..Server Returned: ' + str(revoked_privileges_status) 
                    err_msg = 'Revoke Privileges failed due to: %s' %( str(revoked_privileges_err) )
                    return dict(status='failure', error= err_msg)



